from __future__ import annotations
import os
from dataclasses import dataclass
from typing import List
import json

import asyncio
import aiohttp

from livekit import rtc
from livekit.agents import (
    stt,
    utils,
    APIConnectOptions,
    DEFAULT_API_CONNECT_OPTIONS,
    APIStatusError,
)
from livekit.agents.types import (
    NOT_GIVEN,
    NotGivenOr,
)
from .log import logger


@dataclass
class STTOptions:
    api_key: str | None
    language: str | None
    detect_language: bool
    interim_results: bool
    punctuate: bool
    model: str
    max_sentence_silence: int = 500
    sample_rate: int = 16000
    workspace: str | None = None

    # 增加热词表 提供热词识别
    # 参考url 创建 热词表
    # https://help.aliyun.com/zh/model-studio/custom-hot-words?spm=a2c4g.11186623.0.0.1a7c2fc2CeNIxu
    vocabulary_id: str | None = None
    # 过滤语气词
    disfluency_removal_enabled: bool = False
    # 设置是否开启语义断句，默认关闭。
    semantic_punctuation_enabled: bool = False
    # 设置是否开启标点预测，默认关闭。
    punctuation_prediction_enabled: bool = True
    # 设置是否开启文本逆归一化，默认关闭。
    inverse_text_normalization_enabled: bool = True

    def get_ws_url(self):
        return "wss://dashscope.aliyuncs.com/api-ws/v1/inference"

    def get_header(self):
        header = {
            "Authorization": f"bearer {self.api_key}",
            "X-DashScope-DataInspection": "enable",
        }
        if self.workspace is not None:
            header["X-DashScope-WorkSpace"] = self.workspace
        return header

    def get_run_task_params(self, task_id: str):
        params = {
            "header": {
                "action": "run-task",
                "task_id": task_id,
                "streaming": "duplex",
            },
            "payload": {
                "task_group": "audio",
                "task": "asr",
                "function": "recognition",
                "model": self.model,
                "parameters": {
                    "format": "wav",
                    "sample_rate": self.sample_rate,
                    "vocabulary_id": self.vocabulary_id,
                    "disfluency_removal_enabled": self.disfluency_removal_enabled,
                    "semantic_punctuation_enabled": self.semantic_punctuation_enabled,
                    "punctuation_prediction_enabled": self.punctuation_prediction_enabled,
                    "inverse_text_normalization_enabled": self.inverse_text_normalization_enabled,
                    "max_sentence_silence": self.max_sentence_silence,
                    "heartbeat": True,
                    "language_hints": [self.language],
                },
                "input": {},
            },
        }
        return params

    def get_finish_task_params(self, task_id: str):
        params = {
            "header": {
                "action": "finish-task",
                "task_id": task_id,
                "streaming": "duplex",
            },
            "payload": {"input": {}},
        }
        return params


class STT(stt.STT):
    def __init__(
        self,
        *,
        language="zh",
        detect_language: bool = False,
        interim_results: bool = True,
        punctuate: bool = True,
        model: str = "paraformer-realtime-v2",
        api_key: str | None = None,
        max_sentence_silence: int = 500,
        disfluency_removal_enabled: bool = False,
        semantic_punctuation_enabled: bool = False,
        punctuation_prediction_enabled: bool = True,
        inverse_text_normalization_enabled: bool = True,
        vocabulary_id: str | None = None,
        workspace: str | None = None,
        http_session: aiohttp.ClientSession | None = None,
    ) -> None:
        super().__init__(
            capabilities=stt.STTCapabilities(
                streaming=True, interim_results=interim_results
            )
        )
        api_key = api_key or os.environ.get("DASHSCOPE_API_KEY")
        if api_key is None:
            raise ValueError("DASHSCOPE API key is required")
        self._opts = STTOptions(
            api_key=api_key,
            language=language,
            detect_language=detect_language,
            interim_results=interim_results,
            punctuate=punctuate,
            model=model,
            max_sentence_silence=max_sentence_silence,
            disfluency_removal_enabled=disfluency_removal_enabled,
            semantic_punctuation_enabled=semantic_punctuation_enabled,
            punctuation_prediction_enabled=punctuation_prediction_enabled,
            inverse_text_normalization_enabled=inverse_text_normalization_enabled,
            vocabulary_id=vocabulary_id,
            workspace=workspace,
        )

        self._session = http_session

    def _ensure_session(self) -> aiohttp.ClientSession:
        if not self._session:
            self._session = utils.http_context.http_session()

        return self._session

    async def _recognize_impl(
        self,
        buffer: utils.AudioBuffer,
        *,
        language: NotGivenOr[str] = NOT_GIVEN,
        conn_options: APIConnectOptions,
    ) -> stt.SpeechEvent:
        raise NotImplementedError("not implemented")

    def stream(
        self,
        *,
        language: str | None = None,
        conn_options: APIConnectOptions = DEFAULT_API_CONNECT_OPTIONS,
    ) -> "SpeechStream":
        return SpeechStream(
            stt=self,
            opts=self._opts,
            conn_options=conn_options,
            http_session=self._ensure_session(),
        )


class SpeechStream(stt.SpeechStream):
    def __init__(
        self,
        stt: STT,
        opts: STTOptions,
        conn_options: APIConnectOptions,
        http_session: aiohttp.ClientSession,
    ) -> None:
        super().__init__(stt=stt, conn_options=conn_options)

        if opts.language is None:
            raise ValueError("language detection is not supported in streaming mode")
        self._opts: STTOptions = opts
        self._config = opts
        self._speaking = False
        self._closed = False
        self._request_id = utils.shortuuid()
        self._reconnect_event = asyncio.Event()
        self._session = http_session

    async def _connect_ws(self) -> aiohttp.ClientWebSocketResponse:
        ws = await asyncio.wait_for(
            self._session.ws_connect(
                self._opts.get_ws_url(), headers=self._opts.get_header()
            ),
            self._conn_options.timeout,
        )
        logger.info("connected to stt websocket successfully")
        return ws

    async def _run(self) -> None:
        closing_ws = False
        task_id = utils.shortuuid()

        @utils.log_exceptions(logger=logger)
        async def send_task(ws: aiohttp.ClientWebSocketResponse):
            nonlocal closing_ws
            samples_100ms = self._opts.sample_rate // 10
            audio_bstream = utils.audio.AudioByteStream(
                sample_rate=self._opts.sample_rate,
                num_channels=1,
                samples_per_channel=samples_100ms,
            )

            has_ended = False
            async for data in self._input_ch:
                frames: list[rtc.AudioFrame] = []
                if isinstance(data, rtc.AudioFrame):
                    frames.extend(audio_bstream.write(data.data.tobytes()))
                elif isinstance(data, self._FlushSentinel):
                    frames.extend(audio_bstream.flush())
                    has_ended = True

                for frame in frames:
                    await ws.send_bytes(frame.data.tobytes())

                if has_ended:
                    await ws.send_json(self._opts.get_finish_task_params(task_id))
                    has_ended = False

        @utils.log_exceptions(logger=logger)
        async def recv_task(ws: aiohttp.ClientWebSocketResponse):
            nonlocal closing_ws
            while True:
                msg = await ws.receive()
                if msg.type in (
                    aiohttp.WSMsgType.CLOSED,
                    aiohttp.WSMsgType.CLOSE,
                    aiohttp.WSMsgType.CLOSING,
                ):
                    if closing_ws:  # close is expected, see SpeechStream.aclose
                        return

                    # this will trigger a reconnection, see the _run loop
                    raise APIStatusError(message="connection closed unexpectedly")

                try:
                    self._process_stream_event(json.loads(msg.data))
                except Exception:
                    logger.exception("failed to process message")

        ws: aiohttp.ClientWebSocketResponse | None = None

        while True:
            try:
                ws = await self._connect_ws()
                await ws.send_json(self._opts.get_run_task_params(task_id=task_id))
                tasks = [
                    asyncio.create_task(send_task(ws)),
                    asyncio.create_task(recv_task(ws)),
                ]
                wait_reconnect_task = asyncio.create_task(self._reconnect_event.wait())
                try:
                    done, _ = await asyncio.wait(
                        [asyncio.gather(*tasks), wait_reconnect_task],
                        return_when=asyncio.FIRST_COMPLETED,
                    )  # type: ignore

                    # propagate exceptions from completed tasks
                    for task in done:
                        if task != wait_reconnect_task:
                            task.result()

                    if wait_reconnect_task not in done:
                        break

                    self._reconnect_event.clear()
                finally:
                    await utils.aio.gracefully_cancel(*tasks, wait_reconnect_task)
            finally:
                if ws is not None:
                    await ws.close()

    def _process_stream_event(self, data: dict) -> None:
        event_type = data["header"]["event"]
        if event_type == "result-generated":
            output = data["payload"]["output"]["sentence"]
            is_sentence_end = output["sentence_end"]
            start_time = output["begin_time"]
            end_time = output["end_time"]
            text = output["text"]
            if not self._speaking:
                start_event = stt.SpeechEvent(type=stt.SpeechEventType.START_OF_SPEECH)
                self._event_ch.send_nowait(start_event)
                logger.info("transcription start")
                self._speaking = True
            if text and not is_sentence_end:
                alternatives = [
                    stt.SpeechData(
                        language=self._opts.language,
                        text=text,
                        start_time=start_time,
                        end_time=end_time,
                    )
                ]
                interim_event = stt.SpeechEvent(
                    type=stt.SpeechEventType.INTERIM_TRANSCRIPT,
                    request_id=self._request_id,
                    alternatives=alternatives,
                )
                self._event_ch.send_nowait(interim_event)
            if text and is_sentence_end:
                alternatives = [
                    stt.SpeechData(
                        language=self._opts.language,
                        text=text,
                        start_time=start_time,
                        end_time=end_time,
                    )
                ]
                interim_event = stt.SpeechEvent(
                    type=stt.SpeechEventType.FINAL_TRANSCRIPT,
                    request_id=self._request_id,
                    alternatives=alternatives,
                )
                self._event_ch.send_nowait(interim_event)
                end_event = stt.SpeechEvent(
                    type=stt.SpeechEventType.END_OF_SPEECH, request_id=self._request_id
                )
                self._event_ch.send_nowait(end_event)
                self._speaking = False
                logger.info(
                    "transcription end",
                    extra={
                        "text": text,
                        "start_time": start_time,
                        "end_time": end_time,
                    },
                )


def live_transcription_to_speech_data(
    language: str,
    data,
) -> List[stt.SpeechData]:
    return [
        stt.SpeechData(
            language=language,
            start_time=data["begin_time"],
            end_time=data["end_time"],
            confidence=0.0,
            text=data["text"],
        )
    ]



import os
from dataclasses import dataclass
from typing import AsyncIterable, Optional, Dict
import time
import aiohttp
import asyncio
import json

from livekit.agents import tts, APIConnectOptions, DEFAULT_API_CONNECT_OPTIONS, utils
from osc_data.text_stream import TextStreamSentencizer

from .log import logger


STREAM_EOS = "EOS"


@dataclass
class TTSOptions:
    api_key: str
    model: str
    # 语速，取值范围：0.5~2。
    rate: float
    # 音色
    voice: str
    # 合成音频的语速，取值范围：0.5~2。
    speech_rate: int
    # 合成音频的音量，取值范围：0~100。
    volume: int
    # 采样率，取值范围：8000, 16000, 22050, 24000, 44100, 48000
    sample_rate: int
    # 音调，取值范围：0.5~2。
    pitch: float = 1.0

    def get_ws_url(self) -> str:
        return "wss://dashscope.aliyuncs.com/api-ws/v1/inference"

    def get_ws_header(self) -> Dict[str, str]:
        return {
            "Authorization": f"bearer {self.api_key}",
            "X-DashScope-DataInspection": "enable",
        }

    def get_run_task_params(self) -> Dict[str, str]:
        params = {
            "header": {
                "action": "run-task",
                "task_id": utils.shortuuid(),
                "streaming": "duplex",
            },
            "payload": {
                "task_group": "audio",
                "task": "tts",
                "function": "SpeechSynthesizer",
                "model": self.model,
                "parameters": {
                    "text_type": "PlainText",
                    "voice": self.voice,
                    "format": "pcm",
                    "sample_rate": self.sample_rate,
                    "volume": self.volume,
                    "rate": self.rate,
                    "pitch": self.pitch,
                },
                "input": {},
            },
        }
        return params

    def get_continue_task_params(self, text: str) -> Dict[str, str]:
        params = {
            "header": {
                "action": "continue-task",
                "task_id": utils.shortuuid(),
                "streaming": "duplex",
            },
            "payload": {
                "input": {
                    "text": text,
                }
            },
        }
        return params

    def get_finish_task_params(self) -> Dict[str, str]:
        params = {
            "header": {
                "action": "finish-task",
                "task_id": utils.shortuuid(),
                "streaming": "duplex",
            },
            "payload": {"input": {}},
        }
        return params


class TTS(tts.TTS):
    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        sample_rate: int = 24000,
        voice: str = "longcheng",
        model: str = "cosyvoice-v2",
        speech_rate: int = 1,
        volume: int = 100,
        rate: float = 1.0,
        pitch: float = 1.0,
        http_session: aiohttp.ClientSession | None = None,
        max_session_duration: float = 600,
    ) -> None:
        super().__init__(
            capabilities=tts.TTSCapabilities(streaming=True),
            sample_rate=sample_rate,
            num_channels=1,
        )
        api_key = api_key or os.environ.get("DASHSCOPE_API_KEY")
        if not api_key:
            raise ValueError("DASHSCOPE_API_KEY must be set")
        self._session = http_session
        self._opts = TTSOptions(
            model=model,
            api_key=api_key,
            voice=voice,
            speech_rate=speech_rate,
            volume=volume,
            sample_rate=sample_rate,
            rate=rate,
            pitch=pitch,
        )
        self._pool = utils.ConnectionPool[aiohttp.ClientWebSocketResponse](
            connect_cb=self._connect_ws,
            close_cb=self._close_ws,
            max_session_duration=max_session_duration,
            mark_refreshed_on_get=True,
        )

    def _ensure_session(self) -> aiohttp.ClientSession:
        if self._session is None:
            self._session = utils.http_context.http_session()

        return self._session

    async def _connect_ws(self, timeout: float) -> aiohttp.ClientWebSocketResponse:
        session = self._ensure_session()
        url = self._opts.get_ws_url()
        headers = self._opts.get_ws_header()
        return await asyncio.wait_for(
            session.ws_connect(url, headers=headers),
            timeout=timeout,
        )

    async def _close_ws(self, ws: aiohttp.ClientWebSocketResponse):
        await ws.close()

    def synthesize(
        self,
        text: str,
    ) -> AsyncIterable[tts.SynthesizedAudio]:
        raise NotImplementedError

    def stream(
        self, *, conn_options: APIConnectOptions = DEFAULT_API_CONNECT_OPTIONS
    ) -> "SynthesizeStream":
        return SynthesizeStream(tts=self, opts=self._opts, conn_options=conn_options)


class SynthesizeStream(tts.SynthesizeStream):
    def __init__(
        self,
        *,
        tts: TTS,
        opts: TTSOptions,
        conn_options: APIConnectOptions = DEFAULT_API_CONNECT_OPTIONS,
    ):
        super().__init__(tts=tts, conn_options=conn_options)
        self._opts = opts

    async def _run(self, emitter: tts.AudioEmitter) -> None:
        request_id = utils.shortuuid()
        emitter.initialize(
            request_id=request_id,
            sample_rate=self._opts.sample_rate,
            mime_type="audio/pcm",
            stream=True,
            num_channels=1,
            frame_size_ms=200,
        )

        async def _send_task(sentence: str, ws: aiohttp.ClientWebSocketResponse):
            run_task_params = self._opts.get_run_task_params()
            await ws.send_json(run_task_params)
            continue_task_params = self._opts.get_continue_task_params(text=sentence)
            await ws.send_json(continue_task_params)
            finish_task_params = self._opts.get_finish_task_params()
            await ws.send_json(finish_task_params)

        async def _recv_task(ws: aiohttp.ClientWebSocketResponse):
            is_first_response = True
            start_time = time.perf_counter()
            while True:
                try:
                    msg = await ws.receive()
                except Exception as e:
                    logger.warning(f"Error while receiving bytes: {e}")
                    break
                if msg.type == aiohttp.WSMsgType.BINARY:
                    if is_first_response:
                        elapsed_time = time.perf_counter() - start_time
                        logger.info(
                            "tts first response",
                            extra={"spent": round(elapsed_time, 4)},
                        )
                        is_first_response = False
                    emitter.push(data=msg.data)
                elif msg.type == aiohttp.WSMsgType.TEXT:
                    msg_json = json.loads(msg.data)
                    if "header" in msg_json:
                        header = msg_json["header"]
                        if "event" in header:
                            event = header["event"]
                            if event == "task-finished":
                                break
                            if event == "task-failed":
                                logger.error(f"tts task failed: {msg_json}")
                                break

        splitter = TextStreamSentencizer(remove_emoji=True)
        is_first_sentence = True
        start_time = time.perf_counter()
        async for token in self._input_ch:
            if isinstance(token, self._FlushSentinel):
                sentences = splitter.flush()
            else:
                sentences = splitter.push(text=token)
            for sentence in sentences:
                if is_first_sentence:
                    first_sentence_spend = time.perf_counter() - start_time
                    logger.info(
                        "llm first sentence",
                        extra={"spent": str(first_sentence_spend)},
                    )
                    is_first_sentence = False
                logger.info("tts start", extra={"sentence": sentence})
                emitter.start_segment(segment_id=utils.shortuuid())
                async with self._tts._pool.connection(
                    timeout=self._conn_options.timeout
                ) as ws:
                    assert not ws.closed, "WebSocket connection is closed"
                    tasks = [
                        asyncio.create_task(_send_task(sentence=sentence, ws=ws)),
                        asyncio.create_task(_recv_task(ws=ws)),
                    ]
                    await asyncio.gather(*tasks)
                    emitter.end_segment()
                    logger.info("tts end", extra={"sentence": sentence})
                    self._pushed_text = self._pushed_text.replace(sentence, "")
                    await utils.aio.gracefully_cancel(*tasks)
